/*
 *  STMultiMap
 *
 *  Description:
 *    Plastic shader with maps for base color,
 *    diffuse color, specular color, and reflective
 *    color.  The maps can be 1 or 3 channel.  All
 *    maps are indexed using standard st coordinates.
 *
 *  Parameters:
 *    Ka,Kd,Ks,Roughness - standard plastic parameters
 *    Kr - scalar for reflection
 *    ColorMap, DiffuseMap, SpecularMap, ReflectMap -
 *      texture maps for modulating base color,
 *      diffuse color, specular color, and reflected color
 *
 *  History:
 *    March 15, 2001 - written by Scott Iverson
 */


/* returns a color for a single or 3-channel texture map */

color texture1or3(string texturename)
{
   color Ct=color texture(texturename,s,t,"fill",-1);
   if (comp(Ct,1)==-1) {   /* single channel */
     Ct=comp(Ct,0);
     if (comp(Ct,0)==-1) Ct=0;  /* texture map failure */
   }
   return Ct;
}

surface
  STMultiMap (
  float Ka = 1;
  float Kd = .5;
  float Ks = .5;
  float Kr = 1;
  float Roughness = .1;

  string ColorMap="";
  string DiffuseMap="";
  string SpecularMap="";
  string ReflectMap="";
  )

{
    normal Nf = faceforward (normalize(N),I);

    color Cbase=Cs;
    if (ColorMap!="") Cbase=texture1or3(ColorMap);

    color Cdiffuse=1;
    if (DiffuseMap!="") Cdiffuse=texture1or3(DiffuseMap);

    color Cspec=1;
    if ((Ks>0) && (SpecularMap!="")) Cspec=texture1or3(SpecularMap);

    color Crefl=1;
    if ((Kr>0) && (ReflectMap!="")) Crefl=texture1or3(ReflectMap);


    /* ray trace reflections */

    color Cr=0;
    if (Kr>0) {
       vector R=reflect(normalize(I),Nf);
       Cr=Kr*Crefl*trace(P,R);
    }

    Oi = Os;
    Ci = Oi * (Cbase*(Ka*ambient()+Kd*Cdiffuse*diffuse(Nf))+
               Ks*Cspec*specular(Nf,-normalize(I),Roughness)+Cr);
}
