/*
 * metal01.sl -- Surface shader for metal with rust.
 * 
 * Created for BMRT (Blue Moon Rendering Tools)
 *
 * This is one of the first shaders i wroted. It purpose is to simulate rusted metal.
 * It does't do antialiasing in shader so you will propably have to increase pixel samples to
 * achieve satisfying results. Shader is usefull in static pictures, but not for animations.
 *
 * PARAMETERS:
 *    Ka, Kd, Ks, Kr, roughness			The usual
 *    base_frequency				Scale of shaders
 *    bump_depth				Bump size
 *
 * Shader By Toni Bratincevic
*/

#include "reflections.h"
#include "filterwidth.h"
#include "noises.h"

// standard no-antialiased fBm
float fBm (point p; uniform float octaves, lacunarity, gain )
{
   varying float sum = 0, amp = 1;
   varying point pp = p;
   uniform float i;
   
   for (i = 0; i < octaves; i += 1) {
   	sum += amp * noise(pp);
   	amp *= gain; pp *= lacunarity;
   }
   return sum;
}

// header
surface metal01(
   float Ka = 0.1;
   float Kd = 0.33;
   float Ks = 0.78;
   float Kr = 0.2;
   float roughness = 0.55;
   float base_frequency = 0.02;
   float bump_depth = 0.002; 
)
{
   color Ct;
   point Psh = transform ("shader",P); // transforming shading point to shader space 

// BASIC COLOR OF SURFACE
   
   // CLAYER01 - layer that calculates color variations in shader
           
           //first colors
	   //declarations - frequency of pattern in this layer - color 01 - color 02 - color 03 
	   float l1freq  = 20*base_frequency;
	   color l1grey  = color (0.60,0.64,0.65);
	   color l1brown = color (0.86,0.78,0.72);
	   color l2grey = color (0.76,0.78,0.82);
	   color l2brown = color (0.76,0.74,0.68);

	   //calculations
	   float l1_m_var = pow(fBm (Psh*l1freq,4, 4, 0.3 ),2);
	   color layer1_color =
	      color spline (l1_m_var,
	      		    l1grey,
	      		    l1grey,
	      		    l1brown,
	      		    l2grey,
	      		    l2brown,
	      		    l2brown
	      );
           
           // second colors
	   float l2freq  = 200*base_frequency;
	   float l2_m_var = pow(fBm (Psh*l2freq, 5, 10, 0.4 ),4);

	   color layer2_color =
	      color spline (l2_m_var,
	      		    l1brown,
	      		    l1brown,
	      		    l1grey,
	      		    l2brown,
	      		    l2grey,
	      		    l2grey
	      );

      // noise for mixing
      float l3freq  = 20*base_frequency; 
      float l3_m_var = pow(fBm (Psh*l3freq, 5, 10, 0.6 ),5);
      
      //final color of first layer
      Ct = color (mix(layer2_color,layer1_color,l3_m_var)+mix(layer2_color,layer1_color,0.7))/2;



// RUST COLOR - Second color layer is used to calculate rust color and then mix this color with basic surface color

	   //declarations
	   color rust01gr1 = color ( 0.79,0.77,0.62);
	   color rust01gr2 = color ( 0.72,0.70,0.59);
	   color rust01rd1 = color ( 0.86,0.62,0.38);
	   color rust01rd2 = color ( 0.91,0.68,0.28);
	   float l4freq = 3000*base_frequency;

	   float l4_m_var = pow(fBm(Psh*l4freq*base_frequency, 5, 4, 0.5 ),3);
	   color rust_color1 =
	      color spline (l4_m_var,
	      		    rust01gr1,
	      	            rust01gr1,
	      		    rust01gr2,
	      		    rust01rd1,
	      		    rust01rd2,
	      		    rust01rd2
	      );
	   //Ct = rust_color;

           float mix_freq = 20;
           float mix_noise = fBm((Psh+10)*mix_freq*base_frequency,8,5,0.5);
           float color_mix = pow(mix_noise,4);
           float rust_mix_noise = smoothstep(0.85,0.9,mix_noise);

	   color rust_color2 =
	      color spline (color_mix,
	      		    rust01rd2*0.5,
	      		    rust01rd2*0.7,
	      		    rust01rd1*0.9,
	      		    rust01rd1,
	      		    rust01gr2,
	      	            rust01gr1,
	      	            rust01gr1,
	      	            rust01gr1
	      );
           
           color rust_color = mix (rust_color1,rust_color2,0.38);
           Ct = color mix(rust_color,Ct,rust_mix_noise);

   // Kd and Ks variation calculations
       float Pshn_dist = 0.1;
       float Pshn_freq = 600;
       //petturbe Psh point by some fBm noise
       point Pshn = Psh+(Pshn_dist - float fBm(Psh*Pshn_freq*base_frequency+80,3,5,0.5)*(Pshn_dist*2));
       //stripes in x y and z directions
       float KSKD =
          pow(
             ( fBm(Pshn*(100,1,1)*base_frequency,4,10,0.5)+
               fBm(Pshn*(1,100,1)*base_frequency,4,10,0.5)+
               fBm(Pshn*(1,1,100)*base_frequency,4,10,0.5)  )/3
             ,6)*color_mix;
       // Ct = KSKD; // for test
       float Kd_infl = 0.02;
       float Ks_infl = 0.2;
       float Kd1 = (Kd_infl-KSKD*(Kd_infl*2)); 
       float Kr1 = 0.2*(Ks_infl-KSKD*(Ks_infl*2)); 
 	   

   // Bump Mapping
   float bump_noise=
      bump_depth*
      (1-((KSKD+l4_m_var+l3_m_var+l2_m_var)/4)
      +(1*rust_mix_noise));
   float Ks1 = (0.15*(-0.5+rust_mix_noise))+(Ks_infl-bump_noise*(Ks_infl*2)); 
   
   //normal Nf = faceforward(normalize(N),I);
   normal Nf = faceforward(normalize(calculatenormal (P+(bump_noise*faceforward(normalize(N),I)))),I);            
   vector V = -normalize(I);
 
   // Relfection vector 
   vector Refl = reflect(I,normalize(Nf));
   
   // Calculate final in P point of surface
   Ci = Ct*(Ka*ambient() + (Kd+Kd1)*diffuse(Nf)) + Ct*((Ks+Ks1)*specular(Nf,V,roughness));//+(Kr+Kr1)*RayTrace(P,Refl,1,0,0,0,0);  
   Oi = Os;
}