/*VSL 1.0.0*************************************************************
 *
 * VOLUME VSmoke
 *
 * SYNOPSIS: simple volumetric smoke shader
 *
 * AUTHOR: Scott Iverson
 *
 * COPYRIGHT: 2004 SiTex Graphics, Inc.
 *
 * DESCRIPTION:
 *   Simple volumetric smoke shader using ray marching.\n
 *   \n
 *   The UnitSize parameter controls the overall scale of the shaded 
 *   volume.  The UnitSize setting should be roughly 10-100% of the 
 *   overall dimension of the volume in the scene.  By adjusting the 
 *   UnitSize parameter, the same smoke parameters can used for objects 
 *   and scenes of different scales.\n
 *   \n
 *   OpacityPerUnit gives the extent to which the smoke obscures objects 
 *   behind it.\n
 *   \n
 *   ReflectancePerUnit gives the extent to which the smoke reflects light 
 *   toward the viewer.\n
 *   \n
 *   StepsPerUnit gives the number of ray marching steps to perform per 
 *   UnitSize distance.  More steps take longer to render but produce 
 *   better results.  MaxSteps gives the maximum number of steps to march 
 *   for any one shader call.\n
 *   \n
 *   If SmokeOctaves is greater than 0 the smoke will be modulated by some 
 *   Brownian noise to produce an non-uniform effect.\n
 *   \n
 *   This shader is based heavily on the smoke shader in section 12.6 of 
 *   Advanced RenderMan by Apodaca and Gritz.\n
 *
 ***********************************************************************/


//PageSize 4000 3000

//Pos 14 158

/*DESC
  Set global opacity output variable (Oi)
DESC*/
void SetOpacity_0(
  color Opacity; // Os |opacity
)
{
  extern color Oi;
  
  Oi=Opacity;
}

//Pos 29 105

/*DESC
  Set global output color (Ci)
DESC*/
void SetColor_1(
  color Color; // Cs |color
)
{
  extern color Ci;
  
  Ci=Color;
}

//Pos 335 112

/*DESC
  Volumetric smoke function using ray marching.  This code block is 
  based heavily on the smoke shader in section 12.6 of Advanced 
  RenderMan by Apodaca and Gritz.\n
DESC*/
void Smoke_2(
  float UnitSize; // 5 |approximate size of the volume
  float OpacityPerUnit; // 0.5 |smoke density as it blocks light from behind
  float ReflectancePerUnit; // 0.5 |smoke density as it affects light scattering forward
  float StepsPerUnit; // 50 |step size for ray marching
  float MaxSteps; // 200 |maximum number of ray marching steps
  float SmokeVary; // 0 |variation in uniform smoke
  float SmokeOctaves; // 3 |octaves of smoke noise function
  float SmokeFrequencyPerUnit; // 10 |smoke noise frequency
  output color Color; //  |volume color
  output color Opacity; //  |volume opacity
)
{
  float Brownian(point Position;
                 float FilterWidth;
  float Octaves;
  float Lacunarity;
  float Gain;
  )
  {
    float fw=max(FilterWidth,1.0e-6);
  
    float cutoff=min(pow(2,Octaves),0.5/fw);
    float f,fade;
    float g=1;
    float fbm=0;
  
    for (f=1;f<0.5*cutoff;f*=Lacunarity) {
       fbm+=g*(2*float noise(Position*f)-1);
       g*=Gain;
    }
  
    fade=clamp(2*(cutoff-min(f,cutoff))/cutoff,0,1);
    fbm+=g*fade*(2*float noise(Position*f)-1);
    return fbm;
  }
  
  
  void SmokeDensity(point CurP, ShP;
                    uniform float SmokeFreq;
                    uniform float SmokeOctaves;
                    float StepSize, SmokeVary;
                    output color LightScatter;
                    output float Smoke)
  {
    LightScatter = 0;
    illuminance(CurP) {
      extern color Cl;
      float foglight = 1;
      lightsource("__foglight", foglight);
      if (foglight>0) LightScatter += Cl;
    }
    if ((SmokeVary>0) && (SmokeOctaves>0)) {
      point Psmoke = ShP * SmokeFreq;
      Smoke = 0.5 * Brownian(Psmoke, StepSize, SmokeOctaves, 2.0, 0.5);
      Smoke = float smoothstep(-1.0,1.0,Smoke*SmokeVary);
    } else {
      Smoke = 0.5;
    }
  }
  
  extern point P;
  extern vector I;
  
  float LightDensity = ReflectancePerUnit/UnitSize;
  float OpacityDensity = OpacityPerUnit/UnitSize;
  float StepSize = UnitSize/StepsPerUnit;
  float SmokeFrequency = SmokeFrequencyPerUnit/UnitSize;
  
  point CurOrigin = P-I;
  point ShOrigin = transform("shader", CurOrigin);
  
  float dtau, last_dtau;
  color li, last_li;
  
  vector ShIN = normalize( vtransform("shader", I));
  vector CurIN = vtransform("shader", "current", ShIN);
  
  float d = random()*StepSize;
  float End = min(length(I), StepSize*MaxSteps) - 0.0001;
  float ss = min(StepSize, End-d);
  
  Color = 0;
  Opacity = 0;
  
  SmokeDensity(CurOrigin+d*CurIN, ShOrigin + d*ShIN,
               SmokeFrequency, SmokeOctaves, ss, SmokeVary, 
               last_li, last_dtau);
  
  float alfa=0;
  
  while ((d<End)&&(alfa<1)) {
    ss = clamp(ss, 0.005, End-d);
    d += ss;
  
    SmokeDensity(CurOrigin+d*CurIN, ShOrigin + d*ShIN,
                 SmokeFrequency, SmokeOctaves, ss, SmokeVary,
                 li, dtau);
  
    float tau = OpacityDensity * ss * 0.5 * (dtau + last_dtau);
    color lighttau = LightDensity * ss * 0.5 * (li*dtau + last_li*last_dtau);
  
    Color += (1-alfa) * lighttau;
    alfa += (1-alfa) * (1-exp(-tau));
    last_dtau = dtau;
    last_li = li;
  }
  
  Opacity = clamp(alfa,0,1);
  Color = max(Color,0);
}

//Pos 127 106

/*DESC
  composite top color and opacity over bottom color and opacity
DESC*/
void CompOver_3(
  color TopColor; // Cs |top color
  color TopOpacity; // Os |top opacity
  color BottomColor; // Ci |bottom color
  color BottomOpacity; // Oi |bottom opacity
  output color Color; //  |output color
  output color Opacity; //  |output opacity
)
{
  Color = TopColor + (1-TopOpacity)*BottomColor;
  Opacity = TopOpacity + (1-TopOpacity)*BottomOpacity;
}

//Pos 589 111

void SmokeParms_4(
  output float UnitSize; // 5 |approximate size of the volume
  output float OpacityPerUnit; // 0.5 |smoke density as it blocks light from behind
  output float ReflectancePerUnit; // 0.5 |smoke density as it affects light scattering forward
  output float StepsPerUnit; // 50 |step size for ray marching
  output float MaxSteps; // 200 |maximum number of ray marching steps
  output float SmokeVary; // 0 |variation in uniform smoke
  output float SmokeOctaves; // 3 |octaves of smoke noise function
  output float SmokeFrequencyPerUnit; // 10 |smoke noise frequency
)
{
}

//Link 0 0 3 1
//Link 1 0 3 0
//Link 2 0 4 0
//Link 2 1 4 1
//Link 2 2 4 2
//Link 2 3 4 3
//Link 2 4 4 4
//Link 2 5 4 5
//Link 2 6 4 6
//Link 2 7 4 7
//Link 3 0 2 0
//Link 3 1 2 1

volume VSmoke(
  float UnitSize=5;
  float OpacityPerUnit=0.5;
  float ReflectancePerUnit=0.5;
  float StepsPerUnit=50;
  float MaxSteps=200;
  float SmokeVary=0;
  float SmokeOctaves=3;
  float SmokeFrequencyPerUnit=10;
)
 {
  color Color_2;
  color Opacity_2;
  color Color_3;
  color Opacity_3;

  Smoke_2(UnitSize,OpacityPerUnit,ReflectancePerUnit,StepsPerUnit,MaxSteps,SmokeVary,SmokeOctaves,SmokeFrequencyPerUnit,Color_2,Opacity_2);
  CompOver_3(Color_2,Opacity_2,Ci,Oi,Color_3,Opacity_3);
  SetOpacity_0(Opacity_3);
  SetColor_1(Color_3);
 }
