/*
 * spotlight - extended version of standard spotlight shader
 *
 * PARAMETERS
 *   intensity - scalar light intensity
 *   lightcolor - light color
 *
 *   coneangle - half-angle of illumination cone (in radians)
 *
 *   falloff - controls intensity falloff with distance
 *      0 = no falloff with distance
 *      1 = linear falloff with distance
 *      2 = square-law falloff with distance (physically correct)
 *
 *   shadowname - set to "shadow" for ray-traced shadows
 *   shadowbias - offset for shadow ray to prevent incorrect self-shadowing
 *   shadowcolor - the shadow doesn't have to be black
 *
 *   shadowgroups - groups to test with shadow rays
 *   shadowattribute - name of user attribute to query for
 *     per-object shadow groups override
 *
 *   __shadow - stores the result of the shadow query (if any)
 *   __unshadowed_Cl - stores emitted light w/o shadows
 *   __nondiffuse - set to 1 for a highlights-only light
 *   __nonspecular - set to 1 to prevent highlights from this light
 *   __foglight - set to 0 to exclude from volumes
 *
 * HISTORY
 *   01-22-2008:  added shadowgroups and shadowattribute
 *   10-24-2006:  added __channel and __category
 *   09-23-2004:  added __foglight output var
 *   09-14-2004:  added weight parameter to shadow() call
 *   05-18-2004:  changed default shadowsamples from 16 to 4
 *   11-22-2002:  added support for colored shadows
 *   09-12-2002:  new version created by Scott Iverson
 *
 *  (c) 2002-8 SiTex Graphics, Inc.
 */

light spotlight(
   float intensity = 1;
   color lightcolor = 1;
   point from = point "shader" (0, 0, 0);
   point to   = point "shader" (0, 0, 1);
   float coneangle = radians(30);
   float conedeltaangle = radians(5);  /* transition zone */
   float beamdistribution = 2;  /* exponent for intensity falloff across cone */

   float falloff = 2;  /* exponent for intensity falloff with distance */

   string shadowname = "";  /* shadow map name or "raytrace" for traced shadows */
   float shadowbias = 0.01;
   float shadowblur = 0.01;
   float shadowsamples = 4; /* samples or rays */
   color shadowcolor = color(0,0,0);
   output string shadowgroups = "";
   string shadowattribute = "user:shadowgroups";

   string __category = "";

   output varying color __shadow = color(0,0,0);
   output varying color __unshadowed_Cl = color(0,0,0);
   output float __nondiffuse = 0;  /* set to 1 to exclude from diffuse light */
   output float __nonspecular = 0; /* set to 1 to exclude from highlights */
   output float __foglight = 1;    /* set to 0 to exclude from volumes */
   output float __channel = -1;    /* select per-light output channel */
)
{
   float atten, cosangle;
   uniform vector A = (to - from) / length(to - from);
   uniform float cosoutside = cos(coneangle);
   uniform float cosinside  = cos(coneangle - conedeltaangle);

   illuminate (from, A, coneangle) {

      float lenL=length(L);
      cosangle = (L . A)/lenL;

      atten = 1/pow(lenL,falloff);
      atten *= pow(cosangle, beamdistribution);
      atten *= smoothstep(cosoutside, cosinside, cosangle);

      Cl = (intensity * atten);

      color inshadow=0;
      if (shadowname!="") {
         attribute(shadowattribute, shadowgroups);
         if (shadowgroups=="") {
            inshadow = shadow(shadowname, Ps, "samples", shadowsamples,
                           "blur", shadowblur, "bias", shadowbias, "weight", Cl);
         } else {
            if (shadowgroups=="null") {
               inshadow = 0;
            } else {
               inshadow = shadow(shadowname, Ps, "samples", shadowsamples,
                           "blur", shadowblur, "bias", shadowbias, "weight", Cl,
                           "subset", shadowgroups);
            }
         }
         __shadow=inshadow;
      }

      __unshadowed_Cl = Cl * lightcolor;
      Cl *= mix(lightcolor, shadowcolor, inshadow);
   }

}
